#include "Fsm.h"
#include "Arduino.h"
#include "MotorCtl.h"
#include "Prf.h"


Mode_e        mode;               // FSM mode 
unsigned long targetTime;         // Target time for mode transitions
bool          lastEdge;           // Last value of edge
float         targetHdg;
bool          justLocked;
bool          justUnlocked;
bool          lastLocked;
bool          justEdge;

//FSM: digraph FSM {
//FSM: rankdir=LR;
//FSM: fontsize = 10;
//FSM: size="10,10";
//FSX: page="8.5,11";
//FSM: page="10,10";
//FSM: start=1;
//FSM: overlap=scale;
//FSM: splines=true;
//FSM: orientation=portrait;
//FSM: sep=0.5;
//FSM: Start   [label="Start",                                                shape=circle,  fontsize=10];
//FSM: Command [label="Command",                                              shape=octagon, fontsize=10];
//FSM: Stop    [label="Stop|Dir:Brake|Spd:0|Mode:none|Tgt:none",              shape=record,  fontsize=10];
//FSM: S_WFD   [label="Wait For DMP | Dir:Brake|Spd:0|Mode:none|Tgt:20s",     shape=record,  fontsize=10];
//FSM: S_SF    [label="Seek Fwd|Dir:Fwd|Spd:HI|Mode:hdg|Tgt:2s",              shape=record,  fontsize=10];
//FSM: S_SS    [label="Seek Scan|Dir:Pivot|Spd:LO|Mode:open|Tgt:8-10s",       shape=record,  fontsize=10];
//FSM: S_TF    [label="Track Fwd|Dir:Fwd|Spd:HI|Mode:hdg|Tgt:none",           shape=record,  fontsize=10];
//FSM: S_EB    [label="Edge Back|Dir:Rev|Spd:HI|Mode:none|Tgt:3s",            shape=record,  fontsize=10];
//FSM: Start -> S_WFD;
//FSM: Command -> Stop;



#define LO_SPEED 100
#define HI_SPEED 120


//**************************************************************************
// Sets up the rover to wait for DMP to stabilize
//**************************************************************************
void enter_WAIT_FOR_DMP(uint32_t now)
{

  // Turn off motors
  motorsBrake();
  
  pwmSpeed = 0;
  hdgMode = false;

  // Set timer to for a 20 seconds
  targetTime = now + (20L * 1000L * 1000L);

  mode = WAIT_FOR_DMP;
  Serial.println("Enter WAIT_FOR_DMP");

}



//**************************************************************************
// Sets up the rover to stop
//**************************************************************************
void enter_STOP(uint32_t now)
{

  // Start motors forward
  motorBrake(MOTOR_LT);
  motorBrake(MOTOR_RT);
  
  pwmSpeed = 0;
  hdgMode = false;

  mode = STOP;
  Serial.println("Enter STOP");

}


//**************************************************************************
// Sets up the rover to go forward to a new place to look for beacon
//**************************************************************************
void enter_SEEK_FWD(uint32_t now)
{

  // Start motors forward
  motorsFwd();

  pwmSpeed = HI_SPEED;
  hdgMode = true;
  hdgTgt = hdgDeg;

  // Set timer to for a few seconds
  targetTime = now + (2L * 1000L * 1000L);
  
  mode = SEEK_FWD;
  Serial.println("Enter SEEK_FWD");

}

//**************************************************************************
// Sets up robot to spin looking for beacon
//**************************************************************************
void enter_SEEK_SCAN(uint32_t now)
{

  // Start a pivot turn clockwise
  motorCcw(MOTOR_LT);
  motorBrake(MOTOR_RT);

  pwmSpeed = LO_SPEED;
  hdgMode = false;

  // Set timer for a random amount of turn but enough for at least 360
  //targetHdg = hdgDeg + random(360, 360+180);
  targetTime = now + random(8000, 10000) * 1000L;
  
  mode = SEEK_SCAN;
  Serial.println("Enter SEEK_SCAN");
  
}

//**************************************************************************
// Sets up robot to move forwards towards beacon
//**************************************************************************
void enter_TRACK_FWD(uint32_t now)
{

  // Start motors forward
  motorsFwd();

  pwmSpeed = HI_SPEED;
  hdgMode = true;
  hdgTgt = hdgDeg;

  mode = TRACK_FWD;
  Serial.println("Enter TRACK_FWD");

}

//**************************************************************************
// Sets up robot to back away from an edge
//**************************************************************************
void enter_EDGE_BACK(uint32_t now)
{
  
  // Turn motors on in reverse
  motorsRev();

  pwmSpeed = HI_SPEED;
  hdgMode  = false;
  
  // Set timer for a few seconds
  targetTime = now + (3 * 1000L * 1000L);

  mode = EDGE_BACK;
  Serial.println("Enter EDGE_BACK");

}




//**************************************************************************
// Process Finite State Machine
//**************************************************************************
void processFsm(uint32_t now)
{
    bool locked;

    if (verbose == 0x02)
    {
        Serial.print("F,");    Serial.print(mode);
        Serial.println("");
    }

    // Get lock status
    locked = prfLocked();

    if (locked != lastLocked)
    {
      
      if (locked)
      {
        justLocked   = true;
        justUnlocked = false;
        Serial.println("Just locked");
      } else {
        justLocked   = false;
        justUnlocked = true;
        Serial.println("Just unlocked");
      } 
    }
    lastLocked = locked;


    // If the table edge sensor status changed
    if (edge != lastEdge)
    {

      if (edge)
      {
        justEdge  = true;
        Serial.println("Just edge");
      } else {
        justEdge  = false;
        Serial.println("Just Table");
      }
    }
    lastEdge = edge;


    // Switch based on the state of the machine
    switch (mode) 
    {

    // Robot is waiting for DMP to stabilize
    case WAIT_FOR_DMP:

        // Locked    - Ignore
        // Unlocked  - Ignore
        // Edge      - Ignore
        // Table     - Ignore

        // When time expires
        if (now > targetTime)
        {
            Serial.println("DMP stable");
            enter_SEEK_SCAN(now);
            setHdgBias(-hdgDeg);
            //FSM: S_WFD -> S_SS [label="Time",fontsize=10, len=0.5];
        }
        break;

    // Robot is stopped
    case STOP:
    
        // Locked    - Ignore
        // Unlocked  - Ignore
        // Edge      - Ignore
        // Table     - Ignore
        // Target    - Ignore

        break;

    // Move to a new position to look for beacon
    case SEEK_FWD:
    
        // Locked
        if (justLocked)
        {
            enter_TRACK_FWD(now);
            //FSM: S_SF  -> S_TF [label="locked",fontsize=10];
        }
        
        // Unlocked - Not possible
        
        // Edge     
        if (justEdge)
        {
          enter_EDGE_BACK(now);
          //FSM: S_SF  -> S_EB [label="edge",fontsize=10];
        }
        
        // Table    - Not possible
        
        // When time expires
        if (now > targetTime)
        {
            enter_SEEK_SCAN(now);
            //FSM: S_SF -> S_SS [label="Time",fontsize=10];
        }
        break;

    // Spining looking for beacon
    case SEEK_SCAN:
    
        // Locked
        if (justLocked)
        {
            enter_TRACK_FWD(now);
            //FSM: S_SS  -> S_TF [label="locked",fontsize=10];
        }
        
        // Unlocked  - Not possible
        
        // Edge
        if (justEdge)
        {
          enter_EDGE_BACK(now);
          //FSM: S_SS  -> S_EB [label="edge",fontsize=10];
        }
        
        // Table     - Not possible

        // When we get to the target heading
        //if (hdgTurn > targetHdg)
        if (now > targetTime)
        {
            enter_SEEK_FWD(now);
            //FSM: S_SS -> S_SF [label="Angle",fontsize=10];
        }
        break;

    // Tracking the beacon
    case TRACK_FWD:
    
        // Locked  - Not possible
        // Unlocked
        if (justUnlocked)
        {
            enter_SEEK_SCAN(now);
            //FSM: S_TF -> S_SS [label="Unlocked",fontsize=10];
          
        }
        
        // Edge
        if (justEdge)
        {
          enter_EDGE_BACK(now);
          //FSM: S_TF  -> S_EB [label="Edge",fontsize=10];
        }
        
        // Table  - Not possible
        
        // Target - No target. Just go straight as we are heading towards the beacon
        break;

    // Backing away from table edge
    case EDGE_BACK:
    
        // Locked    - Ignore, we need to get away from edge
        
        // Unlocked  - Ignore, we nees to get away from edge
        
        // Edge      - Still seeing the edge!
        
        // Table     - Ignore, we back up a set amount
        
        // When time expires
        if (now >= targetTime) 
        {
            enter_SEEK_SCAN(now);
            //FSM: S_EB -> S_SS [label="Time",fontsize=10];
        }

        break;

    // Something went wrong
    default:
        // Locked    - Ignore
        // Unlocked  - Ignore
        // Edge      - Ignore
        // Table     - Ignore
        // Target    - Force a stop
        enter_STOP(now);
        break;
    }

}

//FSM: fontsize  = 14;
//FSM: label     = "Rover State Machine";
//FSM: }

