#include "Prf.h"
#include "Arduino.h"

static uint32_t      lockCnt = 0;
static uint32_t      gateLo = 0;
static uint32_t      gateHi = 0;
static uint32_t      gate   = 0;
static uint32_t      fuzz   = 1000L;
static uint32_t      lockedTime = 0;
static uint32_t      codeTime;
static uint32_t      lastPulseTime = 0;
static void          (*prfCb)(bool mode) = 0;
static uint32_t      codeTimes[] =
{
  131200L,
  147000L,
  153000L,
  156000L,
  159000L,
  162000L,
  165000L
};

static void prfIrq(void);

//*************************************************************************
// PRF initialization
//*************************************************************************
void prfInit(int irqPin, void (*fx)(bool mode))
{
  // Define Beacon IRQ pin
  pinMode(irqPin, INPUT);

  // Connect external interrupt pin to ISR
  attachInterrupt(digitalPinToInterrupt(irqPin), prfIrq, FALLING );

  // Save the callback
  prfCb = fx;
  
}


//*************************************************************************
// ISR that handles the interrupt from the IR receiver
//*************************************************************************
static void prfIrq(void)
{
  unsigned long now;
  
  // Get the time
  now = micros();

  lastPulseTime = now;

  // If an early pulse
  if (now < gateLo) 
  {
    // Ignore the pulse, this is noise or another beacon
  }
  
  // If pulse is within the gate
  else if ( (now >= gateLo) && (now <= gateHi) )
  {
    
    if ((lockCnt == 0) && (prfCb != NULL))
    {
        // Call mode change callback
        prfCb(true);
    }
    
    // Increment the lock count
    lockCnt++;

    // Save the time we saw this last pulse
    lockedTime = now;
    
    // Compute the next gate
    gate   = now + codeTime;
    gateLo = gate - fuzz;
    gateHi = gate + fuzz;

  } 

  // If this is a late pulse
  else
  {
    // We got a pulse, but it was after we expected it

    if ((lockCnt != 0) && (prfCb != NULL))
    {
        // Call mode change callback
        prfCb(false);
    }

    // Clear the lock counter
    lockCnt    = 0;
    lockedTime = 0;

    // Calculate a new gate based on this being a good pulse
    gate   = now + codeTime;
    gateLo = gate - fuzz;
    gateHi = gate + fuzz;

  } 
  
}

//*************************************************************************
//
//*************************************************************************
bool prfCheck(uint32_t now)
{

  // If we have not gotten a pulse in two code durations
  if (now > lockedTime + 2*codeTime)
  {

    if ((lockCnt != 0) && (prfCb != NULL))
    {
        // Call mode change callback
        prfCb(false);
    }
    
    // Clear the locked counters
    lockCnt    = 0;    
    lockedTime = now;

  } 

  return (lockCnt != 0);
}


//*************************************************************************
//
//*************************************************************************
void prfSetCode(uint8_t code)
{
  codeTime = codeTimes[code];
}

//*************************************************************************
//
//*************************************************************************
bool prfLocked(void)
{
  return lockCnt != 0;
}


