

//*************************************************************
//* Rover set up to demonstrate IR Beacon and MPU-6050
//*************************************************************
#include "Wire.h"
#include "I2Cdev.h"
#include "MPU6050_6Axis_MotionApps20.h"
#include "MotorCtl.h"
#include "Prf.h"
#include "Range.h"
#include "Cmd.h"
#include "Fsm.h"

// Local prototypes
void  mpuSetup(void);
void  prfCb(bool locked);
void  setHdgBias(float bias);
void  doPid(uint32_t now);


// Pin definitions
const int prfIrq     = 2;     // IRQ 0
const int mpu6050Irq = 3;     // IRQ 1
const int leftA2     = 4;
const int leftA1     = 5;
const int leftPwm    = 6;
const int rightA2    = 7;
const int rightA1    = 8;
const int rightPwm   = 9;
const int spare1     = 10;
const int spare2     = 11;
const int spare3     = 12;
const int lockedPin  = 13;   // Arduino onboard LED
const int sensorPin  = A0;   // Ultrasonic ranger
const int linePin    = A1;   // Line sensor


// Global variables
int16_t       pwmSpeed = 0;       // Desired PWM speed (0-255)
uint32_t      now;                // Current time in us
uint16_t      verbose = 0x0000;   // Verbose printout flag
bool          edge;               // Table edge seen

// Prf related variables
bool          prfStatusChanged = false;
bool          locked;
uint8_t       code = 0;
uint32_t      pidTime = 0;


// MPU6050 variables
MPU6050       mpu;
volatile uint8_t mpuInterrupt = 0;
uint8_t       devStatus;          // return status after each device operation (0 = success, !0 = error)
uint8_t       mpuIntStatus;       // holds actual interrupt status byte from MPU
bool          dmpReady = false;   // set true if DMP init was successful
uint16_t      packetSize;         // expected DMP packet size (default is 42 bytes)
uint16_t      fifoCount;          // count of all bytes currently in FIFO
uint8_t       fifoBuffer[64];     // FIFO storage buffer
float         ypr[3];             // Yaw//Pich//roll
Quaternion    q;                  // [w, x, y, z]  quaternion container
VectorFloat   gravity;            // [x, y, z]     gravity vector
float         hdgDeg;             // Heading in degrees
float         hdgBias = 0.0;      // Difference between truth and MPU6050 headings
float         hdgTgt = 0.0;       // Target heading in deg
float         turns = 0;
float         hdgTurn = 0;
float         hdgDegLast;

// PID paremeters
float kP = 2.5;
float kI = 0.0;
float kD = 0.0;
float error, errorDot, errorLast;
float P = 0.0;
float I = 0.0;
float D = 0.0;
float PID = 0.0;
bool  hdgMode = true;


//*************************************************************************
// Basic Arduino setup routine. Gets called once on powerup
//*************************************************************************
void setup(void) 
{
  
  // Set up serial port
  Serial.begin(115200);

  Serial.println("");
  Serial.println("");
  Serial.println("Starting");

  now = micros();
  
  // Initialize motor subsystem
  motorInit(leftA1,  leftA2,  leftPwm,
            rightA1, rightA2, rightPwm);


  // Beacon locked LED
  pinMode(lockedPin, OUTPUT);

  pinMode(spare1, OUTPUT);
  pinMode(spare2, OUTPUT);
  pinMode(spare3, OUTPUT);

  // Initialize the PRF subsystem
  prfInit(prfIrq, prfCb);

  // Set PRF to track code 
  prfSetCode(code);
  
  // Start waiting for DMP to stabilize
  enter_WAIT_FOR_DMP(now);

  // Initialiaze the MPU6050 subsystem
  mpuSetup();

  // Initialize the range sensor subsystem
  rangeInit(sensorPin);

}

//*************************************************************************
// This is a callback routine that the beacon will call when the lock status changes
// THis is called in the context of an ISR, so go easy on the code.
//*************************************************************************
void prfCb(bool locked)
{

  // Set flag that PRF locked status changed
  prfStatusChanged = true;

  // Set LED indicator based on lock status
  if (locked)
  {
    digitalWrite(lockedPin, HIGH);
  } else {
    digitalWrite(lockedPin, LOW);
  }
  
}


//*************************************************************************
// Basic Arduino loop routines. Gets called from an infinite loop.
//*************************************************************************
void loop(void) 
{
  float range;

  // Get current time
  now = micros();
   
  // Run command parser all the time
  commandParser();

  // Check for lost PRF signal
  prfCheck(now);

  // Read ultrasonic sensor
  // Set edge when a drop off is seen
  range = rangeRead();
  if (!edge)
  {
    if ((range > 10) || (range == 0.0))
    {
      edge = true;
    }
  } else {
    if (range < 6.0)
    {
      edge = false;
    }
  }
  if (verbose & 0x01)
  {
    Serial.print("R,");    Serial.print(range);
    Serial.print(" E,");   Serial.print(edge);
    Serial.println("");
  }

  // Process the FSM
  processFsm(now);

  // Process the MPU6050 data
  processMpu();

  // Try to control heading
  doPid(now);
    
}




//*************************************************************************
// Called when new Dmp data is available
//*************************************************************************
void dmpDataReady(void)
{
  digitalWrite(spare1, HIGH);
  mpuInterrupt++;
  digitalWrite(spare1, LOW);
}

//*************************************************************************
//
//*************************************************************************
void mpuSetup(void) 
{
    // join I2C bus (I2Cdev library doesn't do this automatically)
    #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
        Wire.begin();
        Wire.setClock(400000); // 400kHz I2C clock. Comment this line if having compilation difficulties
    #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
        Fastwire::setup(400, true);
    #endif

    // initialize device
    Serial.println(F("Initializing I2C devices..."));
    mpu.initialize();
    pinMode(mpu6050Irq, INPUT);

    // verify connection
    Serial.println(F("Testing device connections..."));
    Serial.println(mpu.testConnection() ? F("MPU6050 connection successful") : F("MPU6050 connection failed"));

    // load and configure the DMP
    Serial.println(F("Initializing DMP..."));
    devStatus = mpu.dmpInitialize();

    // supply your own gyro offsets here, scaled for min sensitivity
    mpu.setXGyroOffset(220);
    mpu.setYGyroOffset(76);
    mpu.setZGyroOffset(-85);
    mpu.setZAccelOffset(1788); // 1688 factory default for my test chip

    // make sure it worked (returns 0 if so)
    if (devStatus == 0) 
    {
        // turn on the DMP, now that it's ready
        Serial.println(F("Enabling DMP..."));
        mpu.setDMPEnabled(true);

        // enable Arduino interrupt detection
        Serial.println(F("Enabling interrupt detection..."));
        attachInterrupt(digitalPinToInterrupt(mpu6050Irq), dmpDataReady, RISING);
        mpuIntStatus = mpu.getIntStatus();

        // set our DMP Ready flag so the main loop() function knows it's okay to use it
        Serial.println(F("DMP ready! Waiting for first interrupt..."));
        dmpReady = true;

        // get expected DMP packet size for later comparison
        packetSize = mpu.dmpGetFIFOPacketSize();
        
    } else {
        // ERROR!
        // 1 = initial memory load failed
        // 2 = DMP configuration updates failed
        // (if it's going to break, usually the code will be 1)
        Serial.print(F("DMP Initialization failed (code "));
        Serial.print(devStatus);
        Serial.println(F(")"));
        for (;;) {}
    }

}

//*************************************************************************
//
//*************************************************************************
void processMpu(void) 
{


  // If an interrupt has occured (about every 10ms)
  if (mpuInterrupt > 0)
  {

    if (mpuInterrupt > 1)
    {
      Serial.println("falling behind");
    }

    // Reset interrupt flag
    mpuInterrupt = 0;

    // Get current FIFO count
    fifoCount = mpu.getFIFOCount();

    // Flush FIFO if we do not have the right number of bytes
    if ((fifoCount != 42) && (fifoCount != 128))
    {
        if (verbose & 0x10)
        {
            Serial.print(packetSize); Serial.print(" "); Serial.println(fifoCount);
        }
      
        mpu.resetFIFO();
        return;
    }
  
    // Get INT_STATUS byte
    mpuIntStatus = mpu.getIntStatus();

    // check for overflow (this should never happen unless our code is too inefficient)
    if ((mpuIntStatus & 0x10) || fifoCount == 1024) 
    {
        // reset so we can continue cleanly
        mpu.resetFIFO();
        Serial.println(F("FIFO overflow!"));

    // otherwise, check for DMP data ready interrupt (this should happen frequently)
    } else if (mpuIntStatus & 0x02) 
    {
        // wait for correct available data length, should be a VERY short wait
        while (fifoCount < packetSize)
        {
          digitalWrite(spare2, HIGH);
          fifoCount = mpu.getFIFOCount();
          digitalWrite(spare2, LOW);
        }

        // read a packet from FIFO
        mpu.getFIFOBytes(fifoBuffer, packetSize);
        
        // track FIFO count here in case there is > 1 packet available
        // (this lets us immediately read more without waiting for an interrupt)
        fifoCount -= packetSize;

        // Get quaternions then compute YPR
        mpu.dmpGetQuaternion(&q, fifoBuffer);
        mpu.dmpGetGravity(&gravity, &q);
        mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);

        // Save heading
        hdgDeg = ypr[0] * 180.0/M_PI;

        hdgDeg = hdgDeg + hdgBias;
        if (hdgDeg <   0.0) hdgDeg += 360.0;
        if (hdgDeg > 360.0) hdgDeg -= 360.0;
        if (hdgDeg <   0.0) hdgDeg += 360.0;
        if (hdgDeg > 360.0) hdgDeg -= 360.0;

#define DHDG 20.0
        if ( (hdgDeg < DHDG) && (hdgDegLast > 360-DHDG) )
        {
          turns += 360.0;
        } else if ( (hdgDeg > 360-DHDG) && (hdgDegLast < DHDG) )
        {
          turns -= 360.0;
        }
        hdgDegLast = hdgDeg;
        hdgTurn = hdgDeg + turns;
        
        if (verbose & 0x08)
        {
          Serial.print("hdg,");   Serial.print(hdgDeg);
          Serial.print(" t,");    Serial.print(hdgTurn);
          Serial.println("");
        }
    }
  }
}

//*************************************************************************
// Set the bias between truth and the MPU6050 reported heading
//*************************************************************************
void setHdgBias(float bias)
{
  hdgBias = bias;
}


//*************************************************************************
// Target heading is hdgTgt
// Actual heading is hdgDeg
// if we are to follow a constant heading hdgMode
//*************************************************************************
void doPid(uint32_t now)
{

  if (now > pidTime)
  {

    pidTime = now + 500L * 1000L;

    // Error signal
    error = hdgDeg - hdgTgt;
    if (error < -180.0) error += 360.0;
    if (error > +180.0) error -= 360.0;
  
    // Error Dot term
    errorDot = error - errorLast;
    errorLast = error;
  
    // P term
    P = error * kP;
    if (P < -150.0) P = -150.0;
    if (P > +150.0) P = +150.0;
  
    // I term
    I = I + error * kI;
    if (I < -150.0) I = -150.0;
    if (I > +150.0) I = +150.0;
  
    // D term
    D = errorDot * kD;
    if (D < -150.0) D = -150.0;
    if (D > +150.0) D = +150.0;
  
    PID = P + I + D;
    if (PID < -150.0) PID = -150.0;
    if (PID > +150.0) PID = +150.0;
  
    if (pwmSpeed == 0)
    {
      I   = 0.0;
      PID = 0.0;
    }
      
    // If heading is run open loop (for turns etc)
    if (!hdgMode)
    {

      // Reset intergrator and clear PID correction
      I   = 0;
      PID = 0;
      
    }

    motorSpeed(MOTOR_RT, pwmSpeed+PID);
    motorSpeed(MOTOR_LT, pwmSpeed-PID);
  
    if (verbose & 0x04)
    {
      Serial.print("m,");      Serial.print(hdgMode);
      Serial.print(" h,");     Serial.print(hdgDeg);
      Serial.print(" e,");     Serial.print(error);
      Serial.print(" edot,");  Serial.print(errorDot);
      Serial.print(" P,");     Serial.print(P);
      Serial.print(" I,");     Serial.print(I);
      Serial.print(" D,");     Serial.print(D);
      Serial.print(" PWD,");   Serial.print(pwmSpeed);
      Serial.print(" RT,");    Serial.print(pwmSpeed + PID);
      Serial.print(" LT,");    Serial.print(pwmSpeed - PID);
      Serial.println("");
    }
    
  } // right time to update
  
} // end doPid

// End of file
